/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};

// Global sample memory
gc.samples = gc.samples || {};
gc.calibration = gc.calibration || {};
gc.rawSamples = gc.rawSamples || [];

var flt_func;
var flt_func_series;

/*
 *  Boilerplate code for encoding and decoding data transmitted to / received from TI devices.
 */

/**
 * This code implements the IPacketCodec interface.
 *
 * @constructor
 * @implements gc.databind.IPacketCodec
 */
CustomCodec = function()
{
};
CustomCodec.prototype.rxPktCtr = 0;
CustomCodec.prototype.rxMsgCtr = 0;
CustomCodec.prototype.txPktCtr = 0;
CustomCodec.prototype.txMsgCtr = 0;
CustomCodec.prototype.strRxBuffer = "";
CustomCodec.prototype.strTxBuffer = "";
CustomCodec.prototype.isConnect = false;
CustomCodec.prototype.initComplete = false;

CustomCodec.prototype.rxBuf = [];




/**
 * Encodes data into a packet for sending to the target device.
 *
 * @param target - function to call to pass on the encoded data towards the target
 * @param data - The data object from the GUI that is to be encoded into a packet for sending.
 */
CustomCodec.prototype.encode = function(target,data){
    this.txPktCtr++;
    target(data);

    /* The following example code shows how to decode JSON strings going down to the target
       into a Javascript object.  You would use this if you wanted to intercept data from the GUI
       and rework it before passing it on to the target device.

            this.txMsgCtr++;
            if (this.txMsgCtr % 256){
                console.log("Number of tx messages to target: "+this.txMsgCtr);
            }
            var strToSend = JSON.stringify(jsonObj);
            // send the string to the target
            target(strToSend);
        }
        catch(ex){
            console.log("CustomCodec.encode (to target): exception="+ex);
        }
    }
    */
};

function find_avg(elems, ignoreZero)
{
    if (ignoreZero === undefined) ignoreZero = 0;

    var sum = 0;
    var numEl = 0;
    for (var i = 0; i < elems.length; i++ )
    {
        if (elems[i] === 0 && ignoreZero) continue;
        sum += elems[i];
        numEl += 1;
    }

    var avg = sum/numEl;
    return avg?avg:0;
}

function calibrateClick(e)
{
    console.log("Storing last 40 samples as calibration value for 1m");
    gc.calibration = gc.calibration || {};
    for (var freq in gc.samples)
    {
        gc.calibration[freq] = find_avg(gc.samples[freq].slice(-40).map(function(s){return s.filtered_tof}));
    }
    txt_calib = document.querySelector('#widget_textarea_calibration');
    txt_calib.textContent = JSON.stringify(gc.calibration, null, 4);
}

function clearSamplesClick()
{
    gc.samples = {};
    gc.rawSamples = [];
}

function updateMainGraph(target)
{
    var graph = $('#ti_widget_linegraph_main')[0];
    var numPlotSamples = 80;

    var sampleInfo = {};
    var minLen = Infinity;
    var numFreqs = 0;
    var gotCal = false;

    // Find out which frequencies we have got, and the calibration if any
    for (var freq in gc.samples)
    {
        var cal = gc.calibration[freq] || 0;
        if (cal !== 0) gotCal = true;
        sampleInfo[freq] = sampleInfo[freq] || {};

        sampleInfo[freq].cal = cal;
        sampleInfo[freq].len = gc.samples[freq].length;
        minLen = Math.min(minLen, gc.samples[freq].length);

        numFreqs++;
    }

    // Store the first sample for each freq. If we start at the last one, the
    // result isn't stable, so rather make sure we use only min(numSamples_perFreq)-40
    for (freq in sampleInfo)
    {
        var sliceStart = sampleInfo[freq].sliceStart = (minLen<numPlotSamples/*+16*/)?0:minLen-numPlotSamples/*-16*/;

        sampleInfo[freq].series = gc.samples[freq].slice(sliceStart, sliceStart+numPlotSamples/*+16*/).map(function(s){return s.filtered_tof /*s.average_tof*/});
        sampleInfo[freq].filtseries = gc.samples[freq].slice(sliceStart, sliceStart+numPlotSamples/*+16*/).map(function(s){return s.filtered_tof});

        // if (typeof flt_func_series === "function")
        //     sampleInfo[freq].series = flt_func_series(sampleInfo[freq].series);

        sampleInfo[freq].series = sampleInfo[freq].series;//.slice(16);
    }

    var series = [];
    for (var i = 0; i < Math.min(minLen, numPlotSamples); i++)
    {
        var numEl = 0;
        var sum = 0;
        for (freq in sampleInfo)
        {
            val = sampleInfo[freq].series[i];
            // Do not apply a frequency with sample average == 0.
            if (val === 0) continue;
            sum += val - sampleInfo[freq].cal;
            numEl += 1;
        }

        // Push the averaged value, *meters per tick +1 meter for calibration offset at 1m
        var avg = (sum / numEl);
        if (!isNaN(avg))
            series.push(avg * 6.31 + 1);
    }

    graph.series_0_values = series.join(';');

    // Update the meter
    var curr_dst = find_avg(series.slice(-8));
    $('#ti_tile_meter')[0].value = curr_dst;

}

function download_file(filename, text) {
  var blob = new Blob([text], {type: "text/plain;charset=utf-8"});
  saveAs(blob, filename);
}

function downloadSamplesButtonClick()
{
    var dateString = new Date(Date.now()).toISOString();
    download_file('samples_'+dateString+'.json', JSON.stringify(gc.rawSamples, null, 2));
}

/**
 * Decodes packets from the target device into data objects.  One object for each packet of data.
 * Unless you have chained your custom codec with the CR codec, the packet data is not framed,
 * so this method is responsible for decoding partial packets as well as multiple packets.  Partial packets
 * should not be returned, but deferred until the next decode method call with more raw data to process.  As a
 * result, this method may return 0 or more packets of data on each invocation.
 *
 * @param target -  function to call to pass on the encoded data towards the GUI
 * @param data  - The raw data received that needs to be decoded from packets into objects.
 * @return - true if connected (i.e. valid data received), false if not connected
 */
CustomCodec.prototype.decode = function(target, data) {
    this.isConnected = true;
    this.rxPktCtr++;
    window.myCodec = this;
    magic = [0xef, 0xbe, 0xad, 0xde]; // 0xDEADBEEF

    this.rxData = data;

    if (this.bp === undefined) this.bp = new BufferPack();

    // Append to global buffer
    this.rxBuf = this.rxBuf.concat(data);

    // Function for use in arr.filter() to remove RSSI values if the corresponding TOF value is invalid
    var filterNonZeroByIndex = function(e, i) { return this[i] > 0; };

    do
    {
        // Find magic
        for (var i = 0; i < this.rxBuf.length - magic.length; i++)
        {
            test = this.rxBuf.slice(i, i + magic.length);
            if (test.toString() == magic.toString()) // what?
            {
                this.rxBuf = this.rxBuf.slice(i);
                //.log('Found magic, start of buffer is now' + this.rxBuf.slice(0, 5));
                break;
            }
        }

        // Now magic is at the start. Do we have enough data?
        var frame_length = 384+4+4+2+1; // Data + magic + timestamp + freq + fcs
        if ( this.rxBuf.length >=  frame_length)
        {
            // Check checksum
            frame_bytes = this.rxBuf.slice(0, frame_length);
            fcs = frame_bytes.reduce( function(acc, v) { return acc ^ v} );
            //console.log("FCS:" + fcs);

            if (fcs != 0)
            {
                // Frame check failed. Probably missed bytes due to processing overhead.
                // Take away one byte and go look for the next magic start of frame
                console.log("FCS Check failed.");
                this.rxBuf = this.rxBuf.slice(1);
                continue; // Top of do {}.
            }

            var frame;
            var frame_fmt = '<I(magic)<I(ts)<H(freq)256A(tofs)128A(rssis)B(fcs)';
            // Unpack a frame from binary to json
            frame = this.bp.unpack(frame_fmt, frame_bytes);
            frame.tofs = this.bp.unpack('<128H', frame.tofs);
            frame.rssis = this.bp.unpack('<128b', frame.rssis);
            frame.ts = frame.ts/1e6;

            // Take the parsed part out of the buffer
            //this.rxBuf = this.rxBuf.slice(this.bp.calcLength(frame_fmt, this.rxBuf));
            this.rxBuf = this.rxBuf.slice(frame_length);

            // Global last frame
            this.frame = frame;


            // Store this sample in the sample buffer
            gc.samples[frame.freq] = gc.samples[frame.freq] || [];
            gc.samples[frame.freq].push(frame);
            gc.rawSamples.push(frame);

            // Calculate average and store
            frame.average_tof = find_avg(frame.tofs, 1); // Ignore 0 values as they are invalid for various reasons.

            // See if there is a user-defined filter
            if (typeof gc.filterBurst === "function")
                try {
                    frame.filtered_burst_tof = gc.filterBurst(frame.freq, gc.samples[frame.freq], frame);
                    $('#label_edit_burst_error')[0].textContent = '';
                } catch (err) {
                    console.log(err);
                    $('#label_edit_burst_error')[0].textContent = err.toString();
                    gc.filterBurst = undefined;
                    $('#ti_widget_toggle_filter_burst')[0].checked = false;
                    frame.filtered_burst_tof = frame.average_tof;
                }
            else
                frame.filtered_burst_tof = frame.average_tof;

            // Also filter based on previous samples if avail
            if (typeof gc.filterSeries === "function")
                try {
                    frame.filtered_tof = gc.filterSeries(frame, gc.samples[frame.freq]);
                    $('#label_edit_series_error')[0].textContent = '';
                } catch (err) {
                    console.log(err);
                    $('#label_edit_series_error')[0].textContent = err.toString();
                    gc.filterSeries = undefined;
                    $('#ti_widget_toggle_filter_series')[0].checked = false;
                    frame.filtered_tof = frame.filtered_burst_tof;
                }
            else
                frame.filtered_tof = frame.filtered_burst_tof;

            frame.percentValid = frame.tofs.filter( function(e) { return e>0 }).length / frame.tofs.length;

            // Calculate average rssi also, filtered on valid tof
            frame.average_rssi = find_avg(frame.rssis.filter(filterNonZeroByIndex, frame.tofs));

            // Hacky graphing
            var graph = document.querySelector('#ti_widget_graph_'+ frame.freq.toString());

            if (graph !== undefined)
            {
                // Hack since it doesn't care about setting.
                graph.hideYAxis1 = true; // Hide axis for average
                graph.hideYAxis2 = true; // Hide axis for PER%

                // Get filtered of last 60 samples
                var series = gc.samples[frame.freq].slice(-60).map(function(s){return s.average_tof});
                // Don't display frames of all 0's.
                //series = series.filter((v)=>v>0);

                // Run time filter if enabled
                // if ($('#ti_widget_toggle_filter_series')[0].checked)
                //     var flt_func_series = new Function('values', $('#widget_textarea_filter_series')[0].value);
                // if (typeof flt_func_series === "function")
                //     series = flt_func_series(series);

                // Cut off only last 40.
                series = series.slice(-40);

                // Set plot values
                graph.series_0_values = series.join(';');

                // // Get average of last 40 samples
                // var seriesRssi = gc.samples[frame.freq].slice(-40).map(function(s){return s.average_rssi}).map(Math.abs);
                // // Set plot values
                // graph.series_2_values = seriesRssi.join(';');

                // Find percentValid for last 40 samples.
                var seriesValid = gc.samples[frame.freq].slice(-40).map(function(s){return s.percentValid * 100});
                graph.series_2_values = seriesValid.join(';');
                graph.series_2_min = 0;
                graph.series_2_max = 100;


                // Now find graph total average
                var seriesAverage = find_avg(series, true);

                // Updat text label with average
                document.querySelector('#ti_widget_label_raw_'+ frame.freq.toString()).label = "Average: " + seriesAverage.toFixed(3).toString();

                // Set values for filtered
                //graph.series_1_values = Array(series.length).fill(seriesAverage.toFixed(3)).join(';');
                var filteredSeries = gc.samples[frame.freq].slice(-40).map(function(s){return s.filtered_tof});
                graph.series_1_values = filteredSeries.join(';');

                // Set same bounding box for samples and average, otherwise they won't align
                graph.series_0_min = graph.series_1_min = Math.min.apply(null, series.concat(filteredSeries)) - 0.2;
                graph.series_0_max = graph.series_1_max = Math.max.apply(null, series.concat(filteredSeries)) + 0.2;


            }
        }
    }
    while( this.rxBuf.length >= frame_length )

    updateMainGraph(target);

    /* The following example code shows how to decode JSON strings coming up from the target
       into a Javascript object.  You would use this if you wanted to intercept data from the target
       and rework it before passing it on to the GUI for binding

    var strRxData = data.map(function(value) { return String.fromCharCode(value);} ).join("");
    this.strRxBuffer = this.strRxBuffer+strRxData;
    var strMessage = '';
    while (this.strRxBuffer.indexOf('\n') !== -1) {
        strMessage = this.strRxBuffer.substring(0, this.strRxBuffer.indexOf('\n') + 1);
        this.strRxBuffer = this.strRxBuffer.substring(strMessage.length);
        // Rework the following code if your codec is not receiving Javascript objects from the GUI
        try {
            var jsonObj = JSON.parse(strMessage);
            this.rxMsgCtr++;
            if (this.rxMsgCtr % 256){
                console.log("Number of tx messages from target: "+this.rxMsgCtr);
            }

            // send the received and decoded javascript object to the application where you can use it to bind to widget properties
            target(jsonObj);
        }
        catch(ex){
            console.log("CustomCodec.decode (from target): strMessage="+strMessage+", exception="+ex);
        }
    }
    */

    // Return true to set the 'Hardware Connected' status in the status bar
    return this.isConnected;

};

var close = function(){
    gc.databind.unregisterCustomCodec("custom",CustomCodec);
    CustomCodec.initComplete = false;
    CustomCodec.isConnected = false;
};

function savecalibClick(e)
{
    gc.fileCache.writeJsonFile('app/calibration.json', gc.calibration);
}

// Wait for DOMContentLoaded event before trying to access the application template
var init = function() {
    // You can 'chain' codecs together to make it easier to work with the data
    // coming from the device or going to the device.
    // The following codecs are provided:
    //   CR  - use this if the target data uses /n as a delimiter for framing.
    //         The CR codec will take care of buffering the incoming data so that
    //         your codec will only receive complete 'ready to use' packets
    //  If your target uses /n delimited strings, set baseCodecs = "CR"
    //  If your target uses some other format for data (e.g. binary data or custom strings) leave baseCodecs = "undefined"

    var baseCodecs = undefined;
    console.log("custom codec: about to register..."); // log a message that you will be able to see in the debugger
    gc.databind.registerCustomCodec("custom",CustomCodec, baseCodecs);
    CustomCodec.initComplete = true;


    templateObj = document.querySelector('#template_obj');

    // Wait for the template to fire a dom-change event to indicate that it has been 'stamped'
    // before trying to access components in the application.
	templateObj.addEventListener('dom-change',function(){
	    if (initComplete) return;
	    this.async(function(){

            // Send 'm' to device when connected.
            if (templateObj.$.my_usb_uart.serialPortModel.serialIO.currentState.name.indexOf('running') == 0)
            {
                // Send "m" to entice device into being a ToF master
                templateObj.$.my_usb_uart.sendValue('m');
                console.log("Sending MASTER to device");
            }
            else
            {
                templateObj.$.my_usb_uart.addEventListener('serialPortOpen', function() {
                    var state = templateObj.$.my_usb_uart.serialPortModel.serialIO.currentState.name;
                    console.log('-------============== SERIALPORT APP Callback ====  ');
                    if (state.indexOf('running') === 0)
                    {
                        // Send "m" to entice device into being a ToF master
                        templateObj.$.my_usb_uart.sendValue('m');
                        console.log("Sending MASTER to device, delayed");
                    }
                });
            }

    	    document.querySelector('#ti_widget_button0').addEventListener('click', calibrateClick);
    	    document.querySelector('#ti_widget_button_download').addEventListener('click', downloadSamplesButtonClick);
    	    document.querySelector('#ti_widget_button_clear').addEventListener('click', clearSamplesClick);
            document.querySelector('#ti_widget_button_save_calib').addEventListener('click', savecalibClick);

            txt_calib = document.querySelector('#widget_textarea_calibration');
            gc.fileCache.readTextFile('app/calibration.json')
                .then( function (v) {
                    txt_calib.textContent = v;
                    return v;
                }, console.log('Could not load stored calibration data'))
                .then( function (v) {
                    gc.calibration = JSON.parse(v);
                })


            // Re-evaluate functions on enable
            $('#ti_widget_toggle_filter_burst').on('change', function(e){
                if ($('#ti_widget_toggle_filter_burst')[0].checked)
                {
                    //flt_func = new Function('tofValues', $('#widget_textarea_filter_burst')[0].value);
                    gc.filterBurst = eval(gc.ace_editor_burst.getValue());
                }
                else
                {
                    gc.filterBurst = undefined;
                }
            });

            $('#ti_widget_toggle_filter_series').on('change', function(e){
                if ($('#ti_widget_toggle_filter_series')[0].checked)
                {
                    //flt_func_series = new Function('values', $('#widget_textarea_filter_series')[0].value);
                    gc.filterSeries = eval(gc.ace_editor_time.getValue());
                }
                else
                {
                    gc.filterSeries = undefined;
                }
            });

            // Ace editor
            gc.ace_editor = undefined;
            gc.fileCache.readTextFile('ace-noconflict/ace.js')
                .then(function(v) {
                    eval(v);
                    ace.config.set('basePath', 'ace-noconflict/');

                    var editor_burst = ace.edit('ace_edit_burst');
                    var editor_time = ace.edit('ace_edit_time');
                    gc.ace_editor_time = editor_time;
                    gc.ace_editor_burst = editor_burst;

                    [editor_time, editor_burst].forEach( function(editor) {
                        editor.setTheme("ace/theme/eclipse");
                        editor.getSession().setMode("ace/mode/javascript");
                        editor.getSession().setUseWrapMode(true);
                        editor.getSession().setUseSoftTabs(true);
                        editor.getSession().setNewLineMode('auto');
                    });

                    gc.fileCache.readTextFile('app/filter_burst.js')
                        .then(function(v) { editor_burst.setValue(v); editor_burst.navigateFileStart() });

                    gc.fileCache.readTextFile('app/filter_time.js')
                        .then(function(v) { editor_time.setValue(v); editor_time.navigateFileStart() });

                    editor_burst.on('change', function(){
                        $('#ti_widget_toggle_filter_burst')[0].checked = false;
                    });

                    editor_time.on('change', function(){
                        $('#ti_widget_toggle_filter_series')[0].checked = false;
                    });

                });

        },1);

	});

};

templateObj = document.querySelector('#template_obj');
if (templateObj) {
    init();
} else {
    document.addEventListener('DOMContentLoaded',init.bind(this));
}
